@testset "power" begin

    p = Power(3)
    @test p isa Transform
    @test cardinality(p) == OneToOne()

    # TODO: all of these should be part of some test utils
    @testset "Vector" begin
        x = [1, 2, 3, 4, 5]
        expected = [1, 8, 27, 64, 125]

        @test FeatureTransforms.apply(x, p) == expected
        @test p(x) == expected

        _x = copy(x)
        FeatureTransforms.apply!(_x, p)
        @test _x == expected

        @testset "inds" begin
            @test FeatureTransforms.apply(x, p; inds=2:5) ==  expected[2:5]
            @test FeatureTransforms.apply(x, p; dims=:) == expected
            @test FeatureTransforms.apply(x, p; dims=1) == expected
            @test FeatureTransforms.apply(x, p; dims=1, inds=[2, 3, 4, 5]) == expected[2:5]

            @test_throws BoundsError FeatureTransforms.apply(x, p; dims=2)
        end

        @testset "apply_append" begin
            @test FeatureTransforms.apply_append(x, p, append_dim=1) == vcat(x, expected)
        end
    end

    @testset "Matrix" begin
        M = [1 2 3; 4 5 6]
        expected = [1 8 27; 64 125 216]

        @testset "dims = $d" for d in (Colon(), 1, 2)
            @test FeatureTransforms.apply(M, p; dims=d) == expected
            @test p(M; dims=d) == expected

            _M = copy(M)
            FeatureTransforms.apply!(_M, p; dims=d)
            @test _M == expected
        end

        @testset "inds" begin
            @test FeatureTransforms.apply(M, p; inds=[2, 3]) == expected[[2, 3]]
            @test FeatureTransforms.apply(M, p; dims=:, inds=[2, 3]) == expected[[2, 3]]
            @test FeatureTransforms.apply(M, p; dims=1, inds=[2]) == [64 125 216]
            @test FeatureTransforms.apply(M, p; dims=2, inds=[2]) == reshape([8, 125], 2, 1)
        end

        @testset "apply_append" begin
            @test FeatureTransforms.apply_append(M, p, append_dim=1) == cat(M, expected, dims=1)
            @test FeatureTransforms.apply_append(M, p, append_dim=2) == cat(M, expected, dims=2)
            @test FeatureTransforms.apply_append(M, p, append_dim=3) == cat(M, expected, dims=3)
        end
    end

    @testset "AxisArray" begin
        A = AxisArray([1 2 3; 4 5 6], foo=["a", "b"], bar=["x", "y", "z"])
        expected = [1 8 27; 64 125 216]

        @testset "dims = $d" for d in (Colon(), 1, 2)
            transformed = FeatureTransforms.apply(A, p; dims=d)
            # AxisArray doesn't preserve the type it operates on
            @test transformed isa AbstractArray
            @test transformed == expected
        end

        _A = copy(A)
        FeatureTransforms.apply!(_A, p)
        @test _A isa AxisArray
        @test _A == expected

        @testset "inds" begin
            @test FeatureTransforms.apply(A, p; inds=[2, 3]) == expected[[2, 3]]
            @test FeatureTransforms.apply(A, p; dims=:, inds=[2, 3]) == expected[[2, 3]]
            @test FeatureTransforms.apply(A, p; dims=1, inds=[2]) == [64 125 216]
            @test FeatureTransforms.apply(A, p; dims=2, inds=[2]) == reshape([8, 125], 2, 1)
        end

        @testset "apply_append" begin
            @test FeatureTransforms.apply_append(A, p, append_dim=1) == cat(A, expected, dims=1)
            @test FeatureTransforms.apply_append(A, p, append_dim=2) == cat(A, expected, dims=2)
            @test FeatureTransforms.apply_append(A, p, append_dim=3) == cat(A, expected, dims=3)
        end
    end

    @testset "AxisKey" begin
        A = KeyedArray([1 2 3; 4 5 6], foo=["a", "b"], bar=["x", "y", "z"])
        expected = KeyedArray([1 8 27; 64 125 216], foo=["a", "b"], bar=["x", "y", "z"])

        @testset "dims = $d" for d in (Colon(), :foo, :bar, 1, 2)
            transformed = FeatureTransforms.apply(A, p; dims=d)
            @test transformed isa KeyedArray
            @test transformed == expected
        end

        _A = copy(A)
        FeatureTransforms.apply!(_A, p)
        @test _A isa KeyedArray
        @test _A == expected

        @testset "inds" begin
            @test FeatureTransforms.apply(A, p; inds=[2, 3]) == [64, 8]
            @test FeatureTransforms.apply(A, p; dims=:, inds=[2, 3]) == [64, 8]
            @test FeatureTransforms.apply(A, p; dims=1, inds=[2]) == [64 125 216]
            @test FeatureTransforms.apply(A, p; dims=2, inds=[2]) == reshape([8, 125], 2, 1)
        end

        @testset "apply_append" begin
            expected1 = KeyedArray(
                vcat(A, expected), foo=["a", "b", "a", "b"], bar=["x", "y", "z"]
            )
            @test FeatureTransforms.apply_append(A, p, append_dim=:foo) == expected1
            expected2 = KeyedArray(
                hcat(A, expected), foo=["a", "b"], bar=["x", "y", "z", "x", "y", "z"]
            )
            @test FeatureTransforms.apply_append(A, p, append_dim=:bar) == expected2
            expected3 = KeyedArray(
                cat(A, expected, dims=3), foo=["a", "b"], bar=["x", "y", "z"], baz=Base.OneTo(2),
            )
            @test FeatureTransforms.apply_append(A, p, append_dim=:baz) == expected3
        end
    end

    @testset "NamedTuple" begin
        nt = (a = [1, 2, 3], b = [4, 5, 6])

        @testset "all cols" begin

            expected_nt = (Column1 = [1, 8, 27], Column2 = [64, 125, 216])
            @test FeatureTransforms.apply(nt, p) == expected_nt
            @test p(nt) == expected_nt

            _nt = deepcopy(nt)
            FeatureTransforms.apply!(_nt, p)
            @test _nt isa NamedTuple{(:a, :b)}
            @test _nt == (a = [1, 8, 27], b = [64, 125, 216])
        end

        @testset "custom header" begin
            expected_nt = (x = [1, 8, 27], y = [64, 125, 216])
            @test FeatureTransforms.apply(nt, p; header=[:x, :y]) == expected_nt
            @test p(nt; header=[:x, :y]) == expected_nt
        end

        @testset "cols = $c" for c in (:a, :b)
            expected = getproperty(nt, c) .^3

            @test FeatureTransforms.apply(nt, p; cols=c) == (Column1 = expected, )
            @test p(nt; cols=c) == (Column1 = expected, )

            @testset "mutating" for _c in (c, [c])
                _nt = deepcopy(nt)
                FeatureTransforms.apply!(_nt, p; cols=_c)
                @test _nt == merge(nt, NamedTuple{(c,)}((expected,)))
                @test _nt isa NamedTuple
            end
        end

        @testset "apply_append" begin
            expected = merge(nt, (Column1 = [1, 8, 27], Column2 = [64, 125, 216]))
            @test FeatureTransforms.apply_append(nt, p) == expected
        end
    end

    @testset "DataFrame" begin
        df = DataFrame(:a => [1, 2, 3], :b => [4, 5, 6])

        @testset "all cols" begin
            expected_df = DataFrame(:Column1 => [1, 8, 27], :Column2 => [64, 125, 216])
            @test FeatureTransforms.apply(df, p) == expected_df
            @test p(df) == expected_df

            _df = deepcopy(df)
            FeatureTransforms.apply!(_df, p)
            @test _df isa DataFrame
            @test _df == DataFrame(:a => [1, 8, 27], :b => [64, 125, 216])
        end

        @testset "custom header" begin
            expected_df = DataFrame(:x => [1, 8, 27], :y => [64, 125, 216])
            @test FeatureTransforms.apply(df, p; header=[:x, :y]) == expected_df
            @test p(df; header=[:x, :y]) == expected_df
        end

        @testset "cols = $c" for c in (:a, :b)
            expected = getproperty(df, c) .^ 3
            @test FeatureTransforms.apply(df, p; cols=[c]) == DataFrame(:Column1=>expected)
            @test FeatureTransforms.apply(df, p; cols=c) == DataFrame(:Column1=>expected)
            @test p(df; cols=[c]) == DataFrame(:Column1=>expected)
        end

        @testset "apply_append" begin
            expected = DataFrame(
                :a => df.a, :b => df.b, :Column1 => [1, 8, 27], :Column2 => [64, 125, 216]
            )
            @test FeatureTransforms.apply_append(df, p) == expected
        end
    end
end
